from typeguard import typechecked
from . import _write
from ._write import SantisedWriteQuery


@typechecked
def sanitise_query(sql_query: str) -> SantisedWriteQuery:
    """
    Sanitises a SQL query to be provided to the :py:func:`write.query`

    :param query: The SQL query to be sanitised.

    :return: A SantisedWriteQuery

    :raises InvalidSQL: if the query is not a write operation
    :raises SQLInjection: if a possible SQL Injection attempt has been found

    """
    return _write._sanitise(sql_query)


def query(connection, *, sql_query: str, values: tuple = ()):
    """
    Executes a write operation query against the MSSQL Server.

    :param connection: An open MSSQL Connection obtained via :py:func:`client.initialize`.
    :param sql_query: An SQL query that has been sanitised via :py:func:`write.sanitise_query`.
    :param values: Optional. A tuple of values to replace any ? placeholders in the query.
        This helps to prevent any SQL injection attempts if input is provided from untrusted sources.


    .. warning::
        Any user input should be handled with extreme care as it may contain SQL injection attempts to
        read, update or delete data from the database. We recommend using the '?' placeholders for any user
        input in the query and passing the user input as a tuple to the `values` parameter.

        Example:

        .. code-block:: python

            user_name = ppa.input_text("Please enter your name")

            query = write.sanitise_query("INSERT users VALUES (?))

            with sql_adaptor() as connection:
                user_record = write.query(connection, sql_query=query, values=(user_name))


    """
    return _write._query(connection, query=sql_query, values=values)
