r"""
Client

This module contains actions for running queries & stored procedures on a Microsoft SQL Server.

!!! danger "Running Queries"

    See [Using Placeholders](../using_placeholders) before using the [select](#select), [insert](#insert), & [update](#update) actions.

    Placeholders are not required when using the [stored_procedure](#stored_procedure) action.
"""

import logging
import datetime
import functools
import uuid
from typing import Any, Dict, Callable, List, Optional, Union

from datacaster.classes import CastDataClass

from ppa import ui
from mssql import read, client, write, _client
from mssql.exceptions import InvalidSQLQuery

from .records import MSSQLServer

Record = Dict[str, Any]


logger = logging.getLogger(__name__)


def _mssql(func: Callable) -> Callable:
    @functools.wraps(func)
    def _mssql_command(mssql: MSSQLServer, *args, **kwargs):
        mssql["port"] = mssql.get("port", 1433)
        mssql["instance"] = mssql.get("instance", "MSSQLSERVER")
        mssql["default_database"] = mssql.pop("database")
        with client.initialize(**mssql)() as open_connection:
            return func(open_connection, *args, **kwargs)

    return _mssql_command


@_mssql
def select(mssql: MSSQLServer, query: str, placeholders: Optional[List[Any]] = None) -> List[Record]:
    """Run a SELECT SQL query.

!!! question "Minimum Plugin Version: 2.0.0"

??? info "Input"

    - __mssql:__ an [MSSQLServer](../inputs_outputs/#mssqlserver) dictionary

    - __query:__ the SQL Query to run

    - __placeholders:__ Optional [Placeholders](../using_placeholders) to format in to the SQL query before it is executed.

??? success "Output"

    Either a:

    - List of dictionaries containing the query results

    - Empty list if the query returned no data

??? tip

    For more information regarding the output from this action, see [Running Select Queries](../running_select_queries).

??? example
{% raw %}
    ```YAML linenums="1"
    - mssql.client.select:
        query: SELECT * FROM users WHERE name LIKE '%?%'
        placeholders:
          - "{{ name }}"
      load:
        mssql: mssql_secrets
      save: users
    ```
{% endraw %}
"""
    execute_func = functools.partial(read.query, mssql, sql_query=read.sanitise_query(query))
    if placeholders:
        records = execute_func(values=placeholders)
    else:
        records = execute_func()

    if not records:
        logger.info("no rows were returned by the query")
        return []

    class RecordData(CastDataClass):
        __annotations__ = records[0].__annotations__
        __class_config__ = {
            "cast_functions": {"types": {datetime.datetime: lambda x: x.strftime("%H:%M:%S %d/%m/%Y") if x else None}},
            "always_cast": [key for key, value in records[0].__annotations__.items() if value == datetime.datetime],
        }

    return [RecordData(**record._asdict()).__dict__ for record in records]


@_mssql
def insert(mssql: MSSQLServer, query: str, placeholders: Optional[List[Any]] = None):
    """Run an INSERT SQL query.

!!! question "Minimum Plugin Version: 2.0.0"

??? info "Input"

    - __mssql:__ an [MSSQLServer](../inputs_outputs/#mssqlserver) dictionary

    - __query:__ the SQL Query to run

    - __placeholders:__ Optional [Placeholders](../using_placeholders) to format in to the SQL query before it is executed.

??? success "Output"

    Nothing is outputted by this action.

??? example
{% raw %}
    ```YAML linenums="1"
    - mssql.client.insert:
        query: INSERT INTO users(username, name) VALUES(?, ?)
        placeholders:
          - "{{ username }}"
          - "{{ name }}"
      load:
        mssql: mssql_secrets
    ```
{% endraw %}
"""
    if not query.lower().startswith("insert"):
        raise InvalidSQLQuery("SQL Query must be an insert record query")

    execute_func = functools.partial(write.query, mssql, sql_query=write.sanitise_query(query))
    if placeholders:
        return execute_func(values=placeholders)
    return execute_func()


@_mssql
def update(mssql: MSSQLServer, query: str, placeholders: Optional[List[Any]] = None):
    """Run an UPDATE SQL query.

!!! question "Minimum Plugin Version: 2.0.0"

??? info "Input"

    - __mssql:__ an [MSSQLServer](../inputs_outputs/#mssqlserver) dictionary

    - __query:__ the SQL Query to run

    - __placeholders:__ Optional [Placeholders](../using_placeholders) to format in to the SQL query before it is executed.

??? success "Output"

    Nothing is outputted by this action.

??? example
{% raw %}
    ```YAML linenums="1"
    - mssql.client.update:
        query: UPDATE users SET surname = ? WHERE id = ?
        placeholders:
          - "{{ surname }}"
          - "{{ user_id }}"
      load:
        mssql: mssql_secrets
    ```
{% endraw %}
"""
    if not query.lower().startswith("update"):
        raise InvalidSQLQuery("SQL Query must be an update record query")

    execute_func = functools.partial(write.query, mssql, sql_query=write.sanitise_query(query))
    if placeholders:
        return execute_func(values=placeholders)
    return execute_func()


@_mssql
def stored_procedure(mssql: MSSQLServer, name: str, parameters: Optional[Dict[str, Any]] = None) -> List[Record]:
    """Run a stored procedure.

!!! question "Minimum Plugin Version: 2.0.0"

??? info "Input"

    - __mssql:__ an [MSSQLServer](../inputs_outputs/#mssqlserver) dictionary

    - __name:__ the name of the stored procedure (**must** start with *sp_*)

    - __parameters:__ Optional dictionary of parameters to pass in to the stored procedure

!!! tip "Filtered Column Types"

    Only columns with the following data types are outputted by this action:

    - text (`string`)
    - numbers (`integer` & `float`)
    - boolean (`true` & `false`)
    - null

    Columns such as **SID** are removed as PPA receives them in bytes format, which is not JSON compatible.

??? success "Output"

    A list of dictionaries.

??? example "Example 1 - Without Parameters"
    ```YAML linenums="1"
    - mssql.client.stored_procedure:
        name: sp_helplogins
      load:
        mssql: mssql_secrets
      save: user_logins
    ```

??? example "Example 2 - With Parameters"
    ```YAML linenums="1"
    - mssql.client.stored_procedure:
        name: sp_helpuser
        parameters:
          name_in_db: domain\\example.user
      load:
        mssql: mssql_secrets
      save: user_info
    ```
"""
    if not name.startswith("sp_"):
        raise InvalidSQLQuery(f"SQL Stored Procedure name must start with 'sp_', received {name}.")

    execute_func = functools.partial(_client._execute, mssql)

    if parameters:
        keys = sorted(list(parameters.keys()))
        args_template = ", ".join([f"@{key} = ?" for key in keys])
        query = f"{name} {args_template}"
        values = tuple([parameters[key] for key in keys])
        logger.debug(f"query: {query}")
        logger.debug(f"values: {values}")
        cursor = execute_func(query=query, values=values)
    else:
        logger.debug(f"query: {name}")
        cursor = execute_func(query=name)

    header = cursor.description
    rows = cursor.fetchall()
    cursor.close()

    # Loop through the header & store the indices of any column whose type we can't properly handle.
    #
    # The header is in this format:
    # (('LoginName', <class 'str'>, None, 78, 78, 0, True), ('SID', <class 'bytearray'>, None, 85, 85, 0, True))
    safe_column_indices = []
    safe_types = {type("str"), type(1), type(None), type(1.0), type(True)}
    for i, head in enumerate(header):
        column_name, column_type = list(map(lambda x: head[x], {0, 1}))
        if column_type in safe_types:
            safe_column_indices.append(i)
            continue
        logger.info(f"column '{column_name}' will be filtered out as its type '{column_type}' is not supported")

    filtered_header_names = [header[i][0] for i in safe_column_indices]
    filtered_rows = [[row[i] for i in safe_column_indices] for row in rows]
    return_data = [
        {
            header_name: (
                row[i].strip() if isinstance(row[i], str) else row[i]
            )  # Sometimes the cells have loads of whitespace in them!
            for i, header_name in enumerate(filtered_header_names)
        }
        for row in filtered_rows
    ]
    return return_data


def output_table(text: str, records: Union[List[Record], Record]) -> None:
    """Display a list of records in a table.

!!! question "Minimum Plugin Version: 2.0.0"

??? info "Input"

    - __text:__ the title of the table

    - __groups:__ a single or list of dictionaries

??? success "Output"

    Nothing is outputted by this action.

??? example

    Getting all records from a users table and displaying it:

    ```YAML linenums="1"
    - mssql.client.select:
        query: SELECT * FROM users
      load:
        mssql: mssql_secrets
      save: users

    - mssql.client.output_table:
        text: All Users
      load:
        records: users
    ```
"""
    records = [records] if not isinstance(records, list) else records
    ui.output_table(text, header=list(records[0].keys()), data=[list(record.values()) for record in records])


def input_table(
    text: str, records: Union[List[Record], Record], minimum: Optional[int] = None, maximum: Optional[int] = None
) -> List[Record]:
    """Display a list of records in a table, & allow the task operator to make a selection.

!!! question "Minimum Plugin Version: 2.0.0"

??? info "Input"

    - __text:__ the title of the table

    - __records:__ a list of dictionaries to display in the table

    - __minimum:__ The minimum number of acceptable selections (Optional)

    - __maximum:__ The maximum number of acceptable selections (Optional)

??? success "Output"

    A list of dictionaries.

??? tip

    - If __neither__ a `minimum` _or_ `maximum` is provided, the task operator will be able to submit 0 selections.

    - If `minimum` _or_ `maximum` are provided, the operation will repeat until a valid number of selections is made.

??? example

    Getting all records from a users table, displaying it and waiting for a single selection:

    ```YAML linenums="1"
    - mssql.client.select:
        query: SELECT * FROM users
      load:
        mssql: mssql_secrets
      save: users

    - mssql.client.output_table:
        text: Select a user
        minimum: 1
        maximum: 1
      load:
        records: users
    ```
"""
    records = [records] if not isinstance(records, list) else records
    table_element_id = str(uuid.uuid4())
    error_element_id = str(uuid.uuid4())
    while True:
        indices = ui.input_table(
            text, header=list(records[0].keys()), data=[list(record.values()) for record in records], element_id=table_element_id,
        )
        count = len(indices)
        if minimum and count < minimum:
            ui.output_error(f"Please choose at least {minimum} record(s)", element_id=error_element_id)
            continue
        if maximum and count > maximum:
            ui.output_error(f"Please choose no more than {maximum} record(s)", element_id=error_element_id)
            continue
        break

    return [records[i] for i in indices]
